%% tclp/tclp__functor_types2.pl
%% handler for overloaded function symbols
%% Copyright (C) 2004 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__functor_types2.pl,v 1.2 2004/11/08 15:44:24 coquery Exp $

:- module(tclp__functor_types,[
                               apply/4,
                               applyG/4,
                               multi_functor_reinit/0,
                               try_reduce/0,
                               set_label_functors/1,
                               print_functor_types/1,
                               use_heuristic/1,
                               set_look_ahead/1
                              ]).

:- use_module(library(chr)).
:- use_module(library(lists)).

:- use_module(tclp(tclp__clp_types)).
:- use_module(tclp(tclp__utils)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% CHR handler for functors
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

handler tclp__functor_types2.

constraints
%% constraint representing information on an occurence
overloading_occurence/4,
%% constraint containing the possible types of an overloading occurence
occurence_types/2,
%% constraint indicating that the occurence has been instaciated.
instanciated/1,
%% reduces the number of types of an occurence.
%% The second argument is unified with the number of reductions
reduce_types/2,
%% links two scheme args together: whenever one scheme is instanciated, the
%% then other one is reduced
linked_occurences/2,
%% trigger labelling for the given occurence
label_occurence/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% CHR rules

failure @
  overloading_occurence(Id, _Args, FuncDesc, Location),
  occurence_types(Id,[])
  <=> set_last_error(error(Location,
                           [ 'Can''t find a good type for ', FuncDesc])),
      fail.

instanciate @
  overloading_occurence(Id, Args, FuncDesc, Location),
  occurence_types(Id,[SchemeArgs])
  <=> unify_result_args(Args,SchemeArgs,FuncDesc, Location),
      instanciated(Id).

try_schemes @
  overloading_occurence(Id, Args, FuncDesc, Location) \
  occurence_types(Id, Schemes),
  reduce_types(Id, NbReductions)
  <=> filter_schemes(Args, Schemes, Filtered),
      length(Schemes,N1),
      length(Filtered,N2),
      NbReductions is N1 - N2,
      occurence_types(Id,Filtered).

%% this rule should never apply 
reduce_types_elimination @
  reduce_types(_,_)
  <=> throw('trailing reduce_types/2 constraint in tclp__functor_types2').

labelling @
  overloading_occurence(Id, Args, FuncDesc, Location) \
  occurence_types(Id, Schemes),
  label_occurence(Id)
  <=> member(Sc,Schemes),
      occurence_types(Id, [Sc]).

propagationR @
  instanciated(Id), linked_occurences(Id,OtherId)
  <=> reduce_types(OtherId,_).

propagationL @
  instanciated(Id), linked_occurences(OtherId,Id)
  <=> reduce_types(OtherId,_).

instanciated_elimination @
  instanciated(_) <=> true.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Filtering type schemes

/**
 * tries each type scheme given in argument, the last argument is unified with
 * the list of the schemes for which the unification is successful.
 */
filter_schemes(_, [], [], _, _).
filter_schemes(Args,[Sc|Rest],[Sc|Filtered], FuncDesc, Location) :-
        \+ \+ unify_result_args(Args,Sc, FuncDesc, Location),!,
        filter_schemes(Args,Rest,Filtered,FuncDesc,Location).
filter_schemes(Args,[_|Rest],Filtered,FuncDesc,Location) :-
        filter_schemes(Args,Rest,Filtered,FuncDesc,Location).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Instanciating arguments

/**
 * unifies the arguments and the result. If a failure occurs, sets
 * the last error according to this failure
 */
unify_result_args([Result|Args],[SchemeResult|SchemeArgs],
                  FuncDesc, Location) :-
        (
          Result = SchemeResult
        ->
          unify_args(Args,SchemeArgs,FuncDesc,Location)
        ;
          term_type_down(SchemeResult, ConcreteTerm),
          term_type_up(Result, AbstractTerm),
          FuncDesc = Functor/Nargs,
          set_last_error(error(Location,
                               [ 'Incompatible type : ',
                                 Functor/Nargs,
                                 ' has type ',
                                 ConcreteTerm,
                                 ' but is required to have type ',
                                 AbstractTerm
                               ])),!,
          fail
        ).

/**
 * Unifies the arguments. If a failure accurs, sets
 * the last error according to this failure
 */
unify_args([],[],_,_).
unify_args([A|Args],[ScA|ScArgs],FuncDesc, Location) :-
        (
          A = ScA
        ->
          unify_args(Args, ScArgs, FuncDesc, Location)
        ;
          term_type_up(ScA, ConcreteTerm),
          term_type_down(A, AbstractTermD),
          term_type_up(A, AbstractTermU),
          FuncDesc = Functor/Nargs,
          length(CArgs, Len),
          ArgNum is Nargs - Len,
          set_last_error(error(Location,
                               [ 'Incompatible type in ',
                                 Functor/Nargs,
                                 ' argument n ',
                                 ArgNum,
                                 ' has type ',
                                 AbstractTermD-AbstractTermU,
                                 ' but is required to have type ',
                                 ConcreteTerm
                               ])),!,
          fail
        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handling errors to print
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic last_error/1.

/**
 * [set_last_error(Error)] sets the last error to the [Error].
 */
set_last_error(Error) :-
        retractall(last_error(_)),
        assert(last_error(Error)).

/**
 * [set_last_error(Error)] unifies Error with the last error encountered
 * in functor handling.
 */
get_last_error(Error) :-
        last_error(Error).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Applying a functor
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/**
 * [apply(F/N, ArgsTypes, ResultType, Location)] applies [F/N] to arguments
 * which type is in the list [ArgsTypes], the result beeing unified with
 * [ResultType]. [Location] is used for errors
 */
apply(FunctorSpec, ArgsTypes, ResultType, Location) :-
        functor_schemes(FunctorSpec, Schemes,Location),
        length(ArgsTypes,N),
        length(ArgsForSchemes,N),
        link_list(ArgsTypes,ArgsForSchemes),
        overloading_occurence(Id,ArgsForSchemes,FunctorSpec,Location),
        occurence_types(Id,Schemes),
        (Schemes = [_] -> true ; reduce_types(Id,_)).

/**
 * [applyG(F/N, ArgsTypes, ResultType, Location)] applies [F/N] to arguments
 * which type is in the list [ArgsTypes], the result beeing unified with
 * [ResultType]. Considers that the application is made for the head of a
 * clause. [Location] is used for errors
 */
applyG(FunctorSpec , ArgsTypes, ResultType, Location) :-
        functor_schemesG(FunctorSpec, Schemes,Location),
        length(ArgsTypes,N),
        length(ArgsForSchemes,N),
        link_list(ArgsTypes,ArgsForSchemes),
        overloading_occurence(Id,ArgsForSchemes,FunctorSpec,Location),
        occurence_types(Id,Schemes),!.
        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Type of functors and basic types
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% schemes for bodies

/**
 * [functor_schemes/3] gives a list of schemes for a given functor description.
 * The last argument is used for errors.
 */
%% user defined types
functor_schemes(FuncDesc, Types, _Location) :-
        verbose(debug_functor_types, ['asking for ',FuncDesc]),
        user_functor_schemes(FuncDesc, Types),!,
        verbose(debug_functor_types, ['found ',FuncDesc]).
%% default types
functor_schemes(Functor, Type, Location) :-
        default_functor_schemes(Functor, Type, Location).

/**
 * Default and built-in types
 */
%% built-in schemes for int, float, byte, in_byte, character and in_character
default_functor_schemes(X/0, [[Type]],_) :-
        integer(X), !,
        (
          0 =< X, X < 256 -> builtin_type(byte,Type) ;
          X == -1 -> builtin_type(in_byte,Type) ;
          builtin_type(integer, Type)
        ).
default_functor_schemes(X/0, [[Float]],_) :-
        float(X), !, builtin_type(float, Float).
default_functor_schemes(F/0, [[Type]],_) :- !,
        (
          atom_length(F,1) -> builtin_type(character,Type) ;
          F == 'end_of_file' -> builtin_type(in_character,Type)
        ;
          builtin_type(atom, Type),
          verbose(defaults,[ 'Defaulting to atom for ',F/0])
        ).
default_functor_schemes(F/N, [[ Term | AllTerm]], Location) :- !,
        (
          option(default)
        ->
          builtin_type(term, Term),
          uniform_list(N,Term,AllTerm),
          verbose(defaults,[ 'Defaulting to ',AllTerm,
                             ' -> ',Term,' for ',F/N ])
        ;
          print_env(Location),
          print_error(Location,['No type declared for ','$quoted'(F/N)])
        ).

/**
 * User defined schemes.
 */
user_functor_schemes(FuncDesc, Types) :-
        get_declaration(FuncDesc, Types).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% schemes for heads

/**
 * [functor_schemesG/3] gives a list of generalized schemes for a given functor
 * description. The last argument is used for errors.
 */
%% User defined schemes
functor_schemesG(FuncDesc, Types,_Location) :-
        verbose(debug_functor_types, ['asking for ',FuncDesc]),
        user_functor_schemesG(FuncDesc, Types),
        verbose(debug_functor_types, ['found ',FuncDesc]),!.
%% built-in schemes and default schemes int and float
functor_schemesG(FuncDesc, Types, Location) :-
        default_functor_schemes(FuncDesc, Types, Location).

/**
 * User defined schemes, generalized.
 */
user_functor_schemesG(FuncDesc, Types) :-
        get_declarationG(FuncDesc, Types).
