%% tclp__printer.pl
%% utilities to print information

%% Copyright (C) 2001,2003 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__printer.pl,v 1.2 2004/02/06 16:57:08 coquery Exp $

:- module(tclp__printer,[pprint/2,
                         enable_verbose/1,
                         enable_verbose/2, 
                         disable_verbose/1,
                         verbose/2,
                         verbose/3,
                         chrono/2,
                         chrono/3]).


:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__reader)).

:- use_module(library(lists)).

%% Pretty print of quoted terms
convert_var_names(X,'$VAR'(X)) :-
        atom(X),atom_codes(X,[First|_]),
        (First >= 65, First =< 90 ; char_code('_',First)),!.
convert_var_names(X,X) :- var(X),!.
convert_var_names([],[]) :- !.
convert_var_names([X|Xs],[Y|Ys]) :- !,
        convert_var_names(X,Y),
        convert_var_names(Xs,Ys).
convert_var_names(X,Y) :-
        X =.. [ H | Args ],
        convert_var_names(Args,Args2),
        Y =.. [ H | Args2 ].

%% The second argument is either a list of terms to print or a a simgle term
%% to print.
%% Ends the printing with a new line
%% each new line to print (except the last one) is followed by Prompt
true_print(printer(SoA, PrintPred, _), []) :-
        meta_call(PrintPred,[SoA,'\n',[]]).
true_print(printer(SoA, PrintPred, Prompt), [X|Rest]) :- var(X),!,
        meta_call(PrintPred,[SoA,X,[]]),
        true_print(printer(SoA, PrintPred, Prompt), Rest).
true_print(printer(SoA, PrintPred, Prompt), ['$quoted'(X)|Rest]) :-
        convert_var_names(X,T),
        meta_call(PrintPred,[SoA,T,[quoted(true),numbervars(true)]]),
        true_print(printer(SoA, PrintPred, Prompt), Rest).
true_print(printer(SoA, PrintPred, Prompt),
           ['$write_option'(T,Options)|Rest]):-
        meta_call(PrintPred,[SoA,T,Options]),
        true_print(printer(SoA, PrintPred, Prompt), Rest).
true_print(printer(SoA, PrintPred, Prompt), ['\n'| Rest ]) :- !,
        meta_call(PrintPred,[SoA,'\n',[]]),
        meta_call(PrintPred,[SoA,Prompt,[]]),
        true_print(printer(SoA, PrintPred, Prompt), Rest).
true_print(printer(SoA, PrintPred, Prompt), [T|Rest]) :-
        meta_call(PrintPred,[SoA,T,[]]),
        true_print(printer(SoA, PrintPred, Prompt), Rest).
true_print(printer(SoA, PrintPred, Prompt), T) :-
        meta_call(PrintPred,[SoA,T,[]]),
        true_print(printer(SoA, PrintPred, Prompt), []) .

%% The printing predicate, which is a call to true_print with the good
%% arguments
pprint(printer(Prompt),L) :-
        pprint(printer(user_output,Prompt),L).
pprint(printer(SoA, Prompt), L) :-
        pprint(printer(SoA, write_term, Prompt), L).
pprint(printer(SoA, PrintPred, Prompt), L) :-
        true_print(printer(SoA, PrintPred, Prompt), [Prompt|L]),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Verbose stuff
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% verbose(Level) is true if the level Level of verbose is activated
:- dynamic verbose/1, no_comment/1.

%% verbose(Mode, Message) prints a message is the level of verbose is enabled
%% always is always enabled, printing with comment by default
verbose(Level, Message) :-
        (verbose(Level);Level==always)
        ->
        (no_comment(Level) -> Prompt='' ; Prompt='%% '),
        pprint(printer(Prompt),Message)
        ; 
        true .

verbose(Level,Loc, Message) :-
        line_number(Loc, Line),
        get_filename(Loc, FileName),
        verbose(Level,['in "',FileName,'", line ',Line,' : '|Message]).        

enable_verbose(Level,DoComment) :-
        asserta(verbose(Level)),
        (DoComment -> true ; asserta(no_comment(Level))).

enable_verbose(Level) :-
        enable_verbose(Level,true).

disable_verbose(Level) :-
        ( var(Level), write('%% Removing all verbose levels !!!'), nl; true ),
        retractall(verbose(Level)),
        retractall(no_comment(Level)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Chrono of actions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

build_chrono_message(Name,Display,Message,Message2a,Message2b) :-
        is_list(Name),!,
        append(['Starting '|Name],[' ...'],Message),
        append(Name,[' done in ', Display, ' s'],Message2a),
        append(['failed '|Name],[' in ', Display, ' s'],Message2b).
build_chrono_message(Name,Display,Message,Message2a,Message2b) :-
        Message = ['Starting ', Name, ' ...'],
        Message2a = [Name, ' done in ', Display, ' s'],
        Message2b = ['failed ',Name, ' in ', Display, ' s'].
        
:- meta_predicate chrono(?,:).
chrono(Name, Goal) :-
        build_chrono_message(Name,Display,Message,Message2a,Message2b),
        verbose(timing,Message),
        statistics(runtime,[FromStart,_]),
        catch((call(Goal),Result=true;Result=fail),E,Result=throw(E)),
        statistics(runtime,[FromStart2,_]),
        Elapsed is FromStart2 - FromStart,
        Display is Elapsed / 1000,
        ((Result=fail ; Result=throw(_)) ->
            Message2 = Message2b ; Message2a = Message2),
        verbose(timing,Message2),
        call(Result).

:- meta_predicate chrono(?,?,:).
chrono(VerboseLevel, Name, Goal) :-
        verbose(VerboseLevel) -> chrono(Name,Goal) ; Goal.

