%% tclp__reader.pl
%% Copyright (C) 2001,2004 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__reader.pl,v 1.4 2004/02/06 17:12:59 coquery Exp $

%:- use_module(tclp__def).

:- module(tclp__reader, [read_one_term/4,
                         args_data/3,
                         variable_name/3,
                         line_number/2,
                         get_filename/2,
                         get_location/2,
                         term_data_from_variable_names/2,
                         add_phrase_transformer/1,
                         term_to_location_atom/2,
                         get_location_phrase_id/2]).

:- use_module(tclp__utils).

%% This module provides utilities to read terms
%% The type of data attached to read terms

%% converts a list of elements X=Y to a list of pairs X-Y
convert_to_pairs([],[]).
convert_to_pairs([(X=Y)|L],[X-Y|R]) :- convert_to_pairs(L,R).

%% converts a list of elements X=Y to a list of X
convert_to_var_list([],[]).
convert_to_var_list([(X=_)|L],[X|R]) :- convert_to_var_list(L,R).

%% parse one term from the stream and returns the corresponding term_data
read_one_term(Stream, Term, TermData, FileName) :-
        catch(read_term(Stream, OrigTerm,
                        [ layout(OrigLayout),
                          variable_names(OrigVariableNames),
                          singletons(Singletons),
                          syntax_errors(error)]),
              E,
              throw(tclp__error_in_file(FileName,E))
             ),
        transform_phrase(OrigTerm,OrigLayout,OrigVariableNames,
                         Term,Layout,VariableNames),
        convert_to_pairs(VariableNames, VariableNames2),
        convert_to_var_list(Singletons, Singletons2),
        TermData = term_data(Layout, VariableNames2, Singletons2, FileName,
                             _PhraseIdVar).

%% eq_list(L) unifies all elements of L
eq_list([]).
eq_list([_]).
eq_list([X,X|L]) :- eq_list([X|L]).

%% Wrapper for args_data
user:args_location(TermData,ArgsData) :-
        args_data(TermData,ArgsData,_).

%% term_data corresponding to the sub terms of a given compound term
args_data(term_data([_|ArgsLayout], VariableNames, Singletons, FileName,PIV),
          ArgsData, Length) :-
        length(ArgsLayout, Length),!,
        args_data_list(ArgsLayout,
                       VariableNames, Singletons, FileName, PIV,
                       ArgsData).
args_data(term_data([Line|_], VariableNames, Singletons, FileName,PIV),
          ArgsData, Length) :- !,
        length(ArgsLayout, Length),
        eq_list([Line|ArgsLayout]),
        args_data_list(ArgsLayout,
                       VariableNames, Singletons, FileName,PIV,
                       ArgsData).
args_data(term_data(Line, VariableNames, Singletons, FileName,PIV),
          ArgsData, Length) :- !,
        length(ArgsLayout, Length),
        eq_list([Line|ArgsLayout]),
        args_data_list(ArgsLayout,
                       VariableNames, Singletons, FileName,PIV,
                       ArgsData).
args_data(term_data(_,_,_,_,_),[],0).

%% transforms a list of layouts into a list of term_data using the given
%% VariableNames and Singletons
args_data_list([],_,_,_,_,[]).
args_data_list([Layout | Layouts], VariableNames, Singletons, FileName,PIV,
               [ term_data(Layout, VariableNames,Singletons, FileName,PIV)
               | TermDatas ]) :-
        args_data_list(Layouts, VariableNames, Singletons, FileName, PIV,
                       TermDatas).

%% unify the given variable with the atom corresponding to its name.
variable_name(V,term_data(_,VariableNames,_,_,_), Name) :-
        variable_name2(V, VariableNames, Name),!.
variable_name(_,_,'_').
variable_name2(V, [N-X|_], Name) :-
        V == X , Name = N,!.
variable_name2(V, [_|Names], Name) :-
        variable_name2(V,Names,Name).


%% unify LineNumber with the number of the line in the layout of the term_data
line_number(term_data([LineNumber|_],_,_,_,_), LineNumber).
line_number(term_data(LineNumber,_,_,_,_), LineNumber).

%% unify FileName with the name of the file corresponding to this term data
get_filename(term_data(_,_,_,FileName,_), FileName).

%% the first argument is a term_data and the second argument is an atom
%% describing the file and the line corresponding to this term_data
get_location(TermData,Location) :-
        line_number(TermData,Line),
        get_filename(TermData,FileName),
        tclp__utils:convert_list_to_atom([FileName,' : line ',Line],Location).

%% convert a term into an atom
%% using location or convert_to_atom
term_to_location_atom(Term,Location) :-
        get_location(Term,Location),!.
term_to_location_atom(Term,Atom) :-
        convert_to_atom(Term,Atom),!.
term_to_location_atom(Term,Term).

%% term_data_from_variable_names builds a term_data term with nothing except
%% the variablenames which is given by the first argument
term_data_from_variable_names(Names,term_data(_,Names,_,_,_)).

%% the phrase id of a term_data
get_location_phrase_id(term_data(_Layout, _VNames, _Singletons, _Filename,PIV),
                       PIV).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% On the fly transformation of phrases
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% The dynamic predicate used to transform phrases
:- dynamic phrase_transformer/1.

%% try phrase transformers until one works, or leave the phrase unchanged
transform_phrase(OrigPhrase,OrigLayout,OrigVarNames,
                 NewPhrase,NewLayout,NewVarNames) :-
        phrase_transformer(P),
        meta_call(P,[ OrigPhrase,OrigLayout,OrigVarNames,
                      NewPhrase,NewLayout,NewVarNames ]),!.
transform_phrase(OrigPhrase,OrigLayout,OrigVarNames,
                 OrigPhrase,OrigLayout,OrigVarNames).

%% Handle the addition of transformers
add_phrase_transformer(P) :-
        asserta(phrase_transformer(P)).
